# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from django.conf.urls import url
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import render
from django.urls import reverse
from django.utils.http import urlencode
from django.utils.translation import ugettext_lazy as _

from cms.plugin_base import CMSPluginBase, PluginMenuItem
from cms.plugin_pool import plugin_pool
from cms.utils import get_language_from_request
from cms.utils.urlutils import admin_reverse

from .compat import LTE_CMS_3_4
from .forms import ExportImportForm, PluginExportForm, PluginImportForm


class PluginImporter(CMSPluginBase):
    system = True
    render_plugin = False

    def get_plugin_urls(self):
        urlpatterns = [
            url(r'^export-plugins/$', self.export_plugins_view, name='cms_export_plugins'),
            url(r'^import-plugins/$', self.import_plugins_view, name='cms_import_plugins'),
        ]
        return urlpatterns

    def get_extra_placeholder_menu_items(self, request, placeholder):  # noqa
        # django-cms 3.4 compatibility
        return self.get_extra_placeholder_menu_items(request, placeholder)

    def get_extra_global_plugin_menu_items(self, request, plugin):
        # django-cms 3.4 compatibility
        return self.get_extra_plugin_menu_items(request, plugin)

    @classmethod
    def get_extra_plugin_menu_items(cls, request, plugin):
        if plugin.plugin_type == cls.__name__:
            return

        data = urlencode({
            'language': get_language_from_request(request),
            'plugin': plugin.pk,
        })
        return [
            PluginMenuItem(
                _('Export plugins'),
                admin_reverse('cms_export_plugins') + '?' + data,
                data={},
                action='none',
                attributes={
                    'icon': 'export',
                },
            ),
            PluginMenuItem(
                _('Import plugins'),
                admin_reverse('cms_import_plugins') + '?' + data,
                data={},
                action='modal',
                attributes={
                    'icon': 'import',
                },
            ),
        ]

    @classmethod  # noqa
    def get_extra_placeholder_menu_items(cls, request, placeholder):  # noqa
        data = urlencode({
            'language': get_language_from_request(request),
            'placeholder': placeholder.pk,
        })
        return [
            PluginMenuItem(
                _('Export plugins'),
                admin_reverse('cms_export_plugins') + '?' + data,
                data={},
                action='none',
                attributes={
                    'icon': 'export',
                },
            ),
            PluginMenuItem(
                _('Import plugins'),
                admin_reverse('cms_import_plugins') + '?' + data,
                data={},
                action='modal',
                attributes={
                    'icon': 'import',
                },
            )
        ]

    @classmethod
    def import_plugins_view(cls, request):
        if not request.user.is_staff:
            raise PermissionDenied

        new_form = ExportImportForm(request.GET or None)

        if new_form.is_valid():
            initial_data = new_form.cleaned_data
        else:
            initial_data = None

        if request.method == 'GET' and not new_form.is_valid():
            return HttpResponseBadRequest(_('Form received unexpected values.'))

        import_form = PluginImportForm(
            data=request.POST or None,
            files=request.FILES or None,
            initial=initial_data,
        )

        if not import_form.is_valid():
            opts = cls.model._meta
            context = {
                'form': import_form,
                'has_change_permission': True,
                'opts': opts,
                'root_path': reverse('admin:index'),
                'is_popup': True,
                'app_label': opts.app_label,
                'media': (cls().media + import_form.media),
            }
            return render(request, 'djangocms_transfer/import_plugins.html', context)

        plugin = import_form.cleaned_data.get('plugin')
        language = import_form.cleaned_data['language']

        if plugin:
            root_id = plugin.pk
            placeholder = plugin.placeholder
        else:
            root_id = None
            placeholder = import_form.cleaned_data.get('placeholder')

        if not placeholder:
            # Page placeholders/plugins import
            # TODO: Check permissions
            import_form.run_import()
            return HttpResponse('<div><div class="messagelist"><div class="success"></div></div></div>')

        tree_order = placeholder.get_plugin_tree_order(language, parent_id=root_id)
        # TODO: Check permissions
        import_form.run_import()

        if LTE_CMS_3_4:
            return HttpResponse('<div><div class="messagelist"><div class="success"></div></div></div>')

        if plugin:
            new_plugins = plugin.reload().get_descendants().exclude(pk__in=tree_order)
            return plugin.get_plugin_class_instance().render_close_frame(request, obj=new_plugins[0])

        from cms.toolbar.utils import get_plugin_tree_as_json

        # Placeholder plugins import
        new_plugins = placeholder.get_plugins(language).exclude(pk__in=tree_order)
        data = json.loads(get_plugin_tree_as_json(request, list(new_plugins)))
        data['plugin_order'] = tree_order + ['__COPY__']
        data['target_placeholder_id'] = placeholder.pk
        context = {'structure_data': json.dumps(data)}
        return render(request, 'djangocms_transfer/placeholder_close_frame.html', context)

    @classmethod
    def export_plugins_view(cls, request):
        if not request.user.is_staff:
            raise PermissionDenied

        form = PluginExportForm(request.GET or None)

        if not form.is_valid():
            return HttpResponseBadRequest(_('Form received unexpected values.'))

        # TODO: Check permissions
        filename = form.get_filename()
        response = HttpResponse(form.run_export(), content_type='application/json')
        response['Content-Disposition'] = 'attachment; filename={}'.format(filename)
        return response


plugin_pool.register_plugin(PluginImporter)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import transaction

from cms.models import CMSPlugin
from cms.utils.plugins import reorder_plugins

from .utils import get_plugin_class


@transaction.atomic
def import_plugins(plugins, placeholder, language, root_plugin_id=None):
    source_map = {}
    new_plugins = []

    if root_plugin_id:
        root_plugin = CMSPlugin.objects.get(pk=root_plugin_id)
        source_map[root_plugin_id] = root_plugin
    else:
        root_plugin = None

    tree_order = placeholder.get_plugin_tree_order(language, parent_id=root_plugin_id)

    for archived_plugin in plugins:
        if archived_plugin.parent_id:
            parent = source_map[archived_plugin.parent_id]
        else:
            parent = root_plugin

        if parent and parent.__class__ != CMSPlugin:
            parent = parent.cmsplugin_ptr

        plugin = archived_plugin.restore(
            placeholder=placeholder,
            language=language,
            parent=parent,
        )
        source_map[archived_plugin.pk] = plugin

        if parent == root_plugin:
            tree_order.append(plugin.pk)
        new_plugins.append(plugin)

    for new_plugin in new_plugins:
        plugin_class = get_plugin_class(new_plugin.plugin_type)

        if getattr(plugin_class, '_has_do_post_copy', False):
            # getattr is used for django CMS 3.4 compatibility
            # apps on 3.4 wishing to leverage this callback will need
            # to manually set the _has_do_post_copy attribute.
            plugin_class.do_post_copy(new_plugin, source_map)

    reorder_plugins(
        placeholder,
        parent_id=root_plugin_id,
        language=language,
        order=tree_order,
    )
    placeholder.mark_as_dirty(language, clear_cache=False)


@transaction.atomic
def import_plugins_to_page(placeholders, page, language):
    page_placeholders = page.rescan_placeholders()

    for archived_placeholder in placeholders:
        plugins = archived_placeholder.plugins
        placeholder = page_placeholders.get(archived_placeholder.slot)

        if placeholder and plugins:
            import_plugins(plugins, placeholder, language)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.http import urlencode
from django.utils.translation import ugettext

from cms.api import get_page_draft
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool
from cms.utils.page_permissions import user_can_change_page
from cms.utils.urlutils import admin_reverse

from .compat import GTE_CMS_3_6


@toolbar_pool.register
class PluginImporter(CMSToolbar):
    class Media:
        css = {
            'all': ('djangocms_transfer/css/transfer.css',)
        }

    def populate(self):
        # always use draft if we have a page
        page = get_page_draft(self.request.current_page)

        if not page:
            return

        if not user_can_change_page(self.request.user, page):
            return

        page_menu = self.toolbar.get_menu('page')

        if not page_menu or page_menu.disabled:
            return

        data = urlencode({
            'language': self.current_lang,
            'cms_page': page.pk,
        })

        if GTE_CMS_3_6:
            not_edit_mode = not self.toolbar.toolbar_language
        else:
            not_edit_mode = not self.toolbar.language

        page_menu.add_break('Page menu importer break')
        page_menu.add_link_item(
            ugettext('Export'),
            url=admin_reverse('cms_export_plugins') + '?' + data,
            disabled=not_edit_mode,
        )
        page_menu.add_modal_item(
            ugettext('Import'),
            url=admin_reverse('cms_import_plugins') + '?' + data,
            disabled=not_edit_mode,
            on_close=getattr(self.toolbar, 'request_path', self.request.path),
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from cms.models import CMSPlugin, Page, Placeholder

from .datastructures import ArchivedPlaceholder, ArchivedPlugin
from .exporter import export_page, export_placeholder, export_plugin
from .importer import import_plugins, import_plugins_to_page


def _object_version_data_hook(data, for_page=False):
    if not data:
        return data

    if 'plugins' in data:
        return ArchivedPlaceholder(
            slot=data['placeholder'],
            plugins=data['plugins'],
        )

    if 'plugin_type' in data:
        return ArchivedPlugin(**data)
    return data


def _get_parsed_data(file_obj, for_page=False):
    raw = file_obj.read().decode('utf-8')
    return json.loads(raw, object_hook=_object_version_data_hook)


class ExportImportForm(forms.Form):
    plugin = forms.ModelChoiceField(
        CMSPlugin.objects.all(),
        required=False,
        widget=forms.HiddenInput(),
    )
    placeholder = forms.ModelChoiceField(
        queryset=Placeholder.objects.all(),
        required=False,
        widget=forms.HiddenInput(),
    )
    cms_page = forms.ModelChoiceField(
        queryset=Page.objects.drafts(),
        required=False,
        widget=forms.HiddenInput(),
    )
    language = forms.ChoiceField(
        choices=settings.LANGUAGES,
        required=True,
        widget=forms.HiddenInput(),
    )

    def clean(self):
        if self.errors:
            return self.cleaned_data

        plugin = self.cleaned_data.get('plugin')
        placeholder = self.cleaned_data.get('placeholder')
        cms_page = self.cleaned_data.get('cms_page')

        if not any([plugin, placeholder, cms_page]):
            message = _('A plugin, placeholder or page is required.')
            raise forms.ValidationError(message)

        if cms_page and (plugin or placeholder):
            message = _('Plugins can be imported to pages, plugins or placeholders. Not all three.')
            raise forms.ValidationError(message)

        if placeholder and (cms_page or plugin):
            message = _('Plugins can be imported to pages, plugins or placeholders. Not all three.')
            raise forms.ValidationError(message)

        if plugin and (cms_page or placeholder):
            message = _('Plugins can be imported to pages, plugins or placeholders. Not all three.')
            raise forms.ValidationError(message)

        if plugin:
            plugin_model = plugin.get_plugin_class().model
            plugin_is_bound = plugin_model.objects.filter(cmsplugin_ptr=plugin).exists()
        else:
            plugin_is_bound = False

        if plugin and not plugin_is_bound:
            raise ValidationError('Plugin is unbound.')
        return self.cleaned_data


class PluginExportForm(ExportImportForm):

    def get_filename(self):
        if self.cleaned_data.get('cms_page'):
            return 'cms_page_plugins.json'
        return 'plugins.json'

    def run_export(self):
        data = self.cleaned_data
        language = data['language']
        plugin = data['plugin']
        placeholder = data['placeholder']

        if plugin:
            return export_plugin(plugin.get_bound_plugin())

        if placeholder:
            return export_placeholder(placeholder, language)
        return export_page(data['cms_page'], language)


class PluginImportForm(ExportImportForm):

    import_file = forms.FileField(required=True)

    def clean(self):
        if self.errors:
            return self.cleaned_data

        import_file = self.cleaned_data['import_file']

        try:
            data = _get_parsed_data(import_file)
        except (ValueError, TypeError):
            raise ValidationError('File is not valid')

        first_item = data[0]
        is_placeholder = isinstance(first_item, ArchivedPlaceholder)
        page_import = bool(self.cleaned_data['cms_page'])
        plugins_import = not page_import

        if (is_placeholder and plugins_import) or (page_import and not is_placeholder):
            raise ValidationError('Incorrect json format used.')

        self.cleaned_data['import_data'] = data
        return self.cleaned_data

    def run_import(self):
        data = self.cleaned_data
        language = data['language']
        target_page = data['cms_page']
        target_plugin = data['plugin']
        target_placeholder = data['placeholder']

        if target_plugin:
            target_plugin_id = target_plugin.pk
            target_placeholder = target_plugin.placeholder
        else:
            target_plugin_id = None

        if target_page:
            import_plugins_to_page(
                placeholders=data['import_data'],
                page=target_page,
                language=language,
            )
        else:
            import_plugins(
                plugins=data['import_data'],
                placeholder=target_placeholder,
                language=language,
                root_plugin_id=target_plugin_id,
            )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from collections import defaultdict

from django.core import serializers

from .utils import get_plugin_fields, get_plugin_model


def get_bound_plugins(plugins):
    # COMPAT: CMS<3.5
    # When we drop support for CMS3.4 we can replace all this method by
    # from cms.utils.plugins import get_bound_plugins
    plugin_types_map = defaultdict(list)
    plugin_ids = []
    plugin_lookup = {}

    # make a map of plugin types, needed later for downcasting
    for plugin in plugins:
        plugin_ids.append(plugin.pk)
        plugin_types_map[plugin.plugin_type].append(plugin.pk)

    for plugin_type, pks in plugin_types_map.items():
        plugin_model = get_plugin_model(plugin_type)
        plugin_queryset = plugin_model.objects.filter(pk__in=pks)

        # put them in a map so we can replace the base CMSPlugins with their
        # downcasted versions
        for instance in plugin_queryset.iterator():
            plugin_lookup[instance.pk] = instance

    for plugin in plugins:
        parent_not_available = (not plugin.parent_id or plugin.parent_id not in plugin_ids)
        # The plugin either has no parent or needs to have a non-ghost parent
        valid_parent = (parent_not_available or plugin.parent_id in plugin_lookup)

        if valid_parent and plugin.pk in plugin_lookup:
            yield plugin_lookup[plugin.pk]


def get_plugin_data(plugin, only_meta=False):
    if only_meta:
        custom_data = None
    else:
        plugin_fields = get_plugin_fields(plugin.plugin_type)
        _plugin_data = serializers.serialize('python', (plugin,), fields=plugin_fields)[0]
        custom_data = _plugin_data['fields']

    plugin_data = {
        'pk': plugin.pk,
        'creation_date': plugin.creation_date,
        'position': plugin.position,
        'plugin_type': plugin.plugin_type,
        'parent_id': plugin.parent_id,
        'data': custom_data,
    }
    return plugin_data



# -*- coding: utf-8 -*-
from distutils.version import LooseVersion

import cms


cms_version = LooseVersion(cms.__version__)

LTE_CMS_3_4 = cms_version < LooseVersion('3.5')
LTE_CMS_3_5 = cms_version < LooseVersion('3.6')
GTE_CMS_3_6 = cms_version >= LooseVersion('3.6')
